/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/* viewImageData.c */

/* viewImageData.c,v 1.19 2010/06/08 21:41:21 jba Exp */
/************************DESCRIPTION***********************************
  Image Data
**********************************************************************/

#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "cadef.h"
#include "alarm.h"
#include "alarmString.h"

#include "viewImageCA.h"
#include "viewImageErrDialog.h"
#include "viewImageData.h"

/* # defines */
#define     imageName              ":image"
#define     imageSizeName          ":image.SIZE"
#define     imageBitsName          ":image.BITS"
#define     imageXresName          ":image.XRES"
#define     imageYresName          ":image.YRES"
#define     imageArrayName         ":image.L"
#define     imageUpdateName        ":readImageC"
#define     imageRoiXName          ":roiXStartC"
#define     imageRoiYName          ":roiYStartC"
#define     imageRoiWidthName      ":roiXSizeC"
#define     imageRoiHeightName     ":roiYSizeC"
#define     setRedLutWFName        ":setRedLutWF"
#define     setBlueLutWFName       ":setBlueLutWF"
#define     setGreenLutWFName      ":setGreenLutWF"
#define     MAX_COLORS               256

#if 0
/* debug locals */
static int readImageCount=0;
#endif

/* locals */
static char * imageNamePrefix;
static char name[50];
/* Current view values */
static int imageX=0;
static int imageY=0;
static int imageWidth=0;
static int imageHeight=0;
static int imageMaxValue=0;
static int imageMinValue=0;
static unsigned char *imageArray=0;
static int subRegionOnly=0;
/* Pv array values */
static int imageArrayWidth=0;
static int imageArrayHeight=0;
static int imageArraySize=0;
static int imageArrayBits=0;
static int imageArrayDBR[]={DBR_CHAR,DBR_SHORT};
static int newImageArrayWidth=0;
static int newImageArrayHeight=0;
static int newImageArraySize=0;
static int newImageArrayBits=0;
/* Pv Region of Interest (ROI) values */
static int imageRoiX=0;
static int imageRoiY=0;
static int imageRoiWidth=0;
static int imageRoiHeight=0;
/* Sub region: ROI or ViewRegion */
static int imageSubX=0;
static int imageSubY=0;
static int imageSubWidth=0;
static int imageSubHeight=0;
/* Color table */
static unsigned char *colorRedArray=0;
static unsigned char *colorBlueArray=0;
static unsigned char *colorGreenArray=0;
/* Modified indicators */
static int imageModified=1;
static int roiModified=1;
static int subRegionOnlyModified=1;
static int colorTableModified=1;
/* Channel access */
static int neverConnected=1;
static chid imageChid=NULL;
static chid imageSizeChid=NULL;
static chid imageBitsChid=NULL;
static chid imageXresChid=NULL;
static chid imageYresChid=NULL;
static evid imageSizeEvid=NULL;
static evid imageBitsEvid=NULL;
static evid imageXresEvid=NULL;
static evid imageYresEvid=NULL;
static chid imageUpdateChid=NULL;
static chid imageRoiXChid=NULL;
static chid imageRoiYChid=NULL;
static chid imageRoiWidthChid=NULL;
static chid imageRoiHeightChid=NULL;
static evid imageRoiXEvid=NULL;
static evid imageRoiYEvid=NULL;
static evid imageRoiWidthEvid=NULL;
static evid imageRoiHeightEvid=NULL;
static evid imageEvid=NULL;
static chid *imageArrayChid=NULL;
static chid setRedLutWFChid=NULL;
static chid setBlueLutWFChid=NULL;
static chid setGreenLutWFChid=NULL;
static evid setRedLutWFEvid=NULL;
static evid setBlueLutWFEvid=NULL;
static evid setGreenLutWFEvid=NULL;

/* forward declarations */

/******************************************************
  viewImageDataGetColorRedArray
******************************************************/
unsigned char * viewImageDataGetColorRedArray()
{
   return colorRedArray;
}

/******************************************************
  viewImageDataGetColorBlueArray
******************************************************/
unsigned char * viewImageDataGetColorBlueArray()
{
   return colorBlueArray;
}

/******************************************************
  viewImageDataGetColorGreenArray
******************************************************/
unsigned char * viewImageDataGetColorGreenArray()
{
   return colorGreenArray;
}

/******************************************************
  viewImageDataGetImageArray
******************************************************/
unsigned char * viewImageDataGetImageArray()
{
   return imageArray;
}

/******************************************************
 viewImageDataGetRoi
******************************************************/
void viewImageDataGetRoi(int *proiX,int *proiY,int *proiWidth,int *proiHeight)
{
  *proiX=imageRoiX;
  *proiY=imageRoiY;
  *proiWidth=imageRoiWidth;
  *proiHeight=imageRoiHeight;
  return;
}

/******************************************************
 viewImageDataGetImageX
******************************************************/
int viewImageDataGetImageX()
{
  return imageX;
}

/******************************************************
 viewImageDataGetImageY
******************************************************/
int viewImageDataGetImageY()
{
    return imageY;
}

/******************************************************
  viewImageDataGetImageArrayWidth
******************************************************/
int viewImageDataGetImageArrayWidth()
{
  return imageArrayWidth;
}

/******************************************************
 viewImageDataGetImageArrayHeight
******************************************************/
int viewImageDataGetImageArrayHeight()
{
  return imageArrayHeight;
}

/******************************************************
  viewImageDataGetImageArraySize
******************************************************/
int viewImageDataGetImageArraySize()
{
  return imageArraySize;
}

/******************************************************
  viewImageDataGetImageArrayBits
******************************************************/
int viewImageDataGetImageArrayBits()
{
  return imageArrayBits;
}

/******************************************************
  viewImageDataGetImageWidth
******************************************************/
int viewImageDataGetImageWidth()
{
  return imageWidth;
}

/******************************************************
 viewImageDataGetImageHeight
******************************************************/
int viewImageDataGetImageHeight()
{
  return imageHeight;
}

/******************************************************
  viewImageDataGetImageMinValue
******************************************************/
int viewImageDataGetImageMinValue()
{
   return imageMinValue;
}

/******************************************************
  viewImageDataGetImageMaxValue
******************************************************/
int viewImageDataGetImageMaxValue()
{
   return imageMaxValue;
}

/******************************************************
  viewImageDataColorTableIsModified
******************************************************/
int viewImageDataColorTableIsModified() 
{
   return colorTableModified;
}

/******************************************************
  viewImageDataImageIsModified
******************************************************/
int viewImageDataImageIsModified() 
{
  return imageModified;
}

/******************************************************
  viewImageDataDataIsModified
******************************************************/
int viewImageDataDataIsModified() 
{
  if (imageModified || roiModified || subRegionOnlyModified) return 1;
  else return 0;
}

/******************************************************
  viewImageDataSetColorTableIsModified
******************************************************/
void viewImageDataSetColorTableIsModified(int val) 
{
   colorTableModified = val;
}

/******************************************************
  viewImageDataSetSubRegion
******************************************************/
void viewImageDataSetSubRegion(int tempX,int tempY,int tempWidth,int tempHeight)
{
  imageSubX=tempX;
  imageSubY=tempY;
  imageSubWidth=tempWidth;
  imageSubHeight=tempHeight;
}

/******************************************************
  viewImageDataSetRoi
******************************************************/
void viewImageDataSetRoi(int X,int Y,int Width,int Height)
{
  int newRoiX,newRoiY,newRoiWidth,newRoiHeight;

  if ( viewImageDataRoiIsConnected()){
    newRoiX=X;
    newRoiY=Y;
    newRoiWidth=Width;
    newRoiHeight=Height;
    viewImageCaPut(DBR_LONG,imageRoiXChid,(void *)&newRoiX);
    viewImageCaPut(DBR_LONG,imageRoiYChid,(void *)&newRoiY);
    viewImageCaPut(DBR_LONG,imageRoiWidthChid,(void *)&newRoiWidth);
    viewImageCaPut(DBR_LONG,imageRoiHeightChid,(void *)&newRoiHeight);
    viewImageCaFlushIO();
  } else {
    imageRoiX=X;
    imageRoiY=Y;
    imageRoiWidth=Width;
    imageRoiHeight=Height;
    roiModified = 1;
  }
}

/******************************************************
  viewImageDataSetSubRegionOnly
******************************************************/
void viewImageDataSetSubRegionOnly(int value)
{
  subRegionOnly=value;
  subRegionOnlyModified=1;
}

/******************************************************
  viewImageDataRequestImageUpdate
******************************************************/
void viewImageDataRequestImageUpdate()
{
  int value=1;
#if 0
time_t now;
static time_t prev=0;
static int count;
now=time(NULL);
count++;
if (now > prev) {
 printf(" Requesting image update count=%d  Read data count %d  %s", count,readImageCount,ctime(&now));
 fflush(stdout);
 count=0;
 readImageCount=0;
 prev=now;
}
#endif

#if 0
time_t now;
now=time(NULL);
printf ("CA put requesting image update:  %s",ctime(&now));
#endif

  viewImageCaPut(DBR_LONG,imageUpdateChid,(void *)&value);
  viewImageCaFlushIO();
}
  
/******************************************************
  viewImageDataUpdateImageIsConnected
******************************************************/
int viewImageDataUpdateImageIsConnected()
{
  if (viewImageCaIsConnected(imageUpdateChid)) return 1;
  else return 0;
}

/******************************************************
  viewImageDataRoiIsConnected
******************************************************/
int viewImageDataRoiIsConnected()
{
  if (viewImageCaIsConnected(imageRoiXChid) &&
      viewImageCaIsConnected(imageRoiYChid) &&
      viewImageCaIsConnected(imageRoiWidthChid) &&
      viewImageCaIsConnected(imageRoiHeightChid)) return 1;
  else return 0;
}

/******************************************************
  viewImageDataClearChannels
******************************************************/
void viewImageDataClearChannels()
{
  int i;

  if (neverConnected) return;

  if (imageEvid != NULL) viewImageCaClearEvent(&imageEvid);
  if (imageChid != NULL) viewImageCaClearChannel(&imageChid);
  if (imageUpdateChid != NULL) viewImageCaClearChannel(&imageUpdateChid);
  if (imageRoiXEvid != NULL) viewImageCaClearEvent(&imageRoiXEvid);
  if (imageRoiYEvid != NULL) viewImageCaClearEvent(&imageRoiYEvid);
  if (imageRoiWidthEvid != NULL) viewImageCaClearEvent(&imageRoiWidthEvid);
  if (imageRoiHeightEvid != NULL) viewImageCaClearEvent(&imageRoiHeightEvid);
  if (imageRoiXChid != NULL) viewImageCaClearChannel(&imageRoiXChid);
  if (imageRoiYChid != NULL) viewImageCaClearChannel(&imageRoiYChid);
  if (imageRoiWidthChid != NULL) viewImageCaClearChannel(&imageRoiWidthChid);
  if (imageRoiHeightChid != NULL) viewImageCaClearChannel(&imageRoiHeightChid);
  if (imageArrayChid !=NULL ){
    for (i=0;i<imageArrayHeight;i++) {
      if (imageArrayChid[i] != NULL) viewImageCaClearChannel(&imageArrayChid[i]);
    }
  }
  if (setRedLutWFEvid != NULL) viewImageCaClearEvent(&setRedLutWFEvid);
  if (setBlueLutWFEvid != NULL) viewImageCaClearEvent(&setBlueLutWFEvid);
  if (setGreenLutWFEvid != NULL) viewImageCaClearEvent(&setGreenLutWFEvid);
  if (setRedLutWFChid != NULL) viewImageCaClearChannel(&setRedLutWFChid);
  if (setBlueLutWFChid != NULL) viewImageCaClearChannel(&setBlueLutWFChid);
  if (setGreenLutWFChid != NULL) viewImageCaClearChannel(&setGreenLutWFChid);
  if (imageSizeEvid != NULL) viewImageCaClearEvent(&imageSizeEvid);
  if (imageBitsEvid != NULL) viewImageCaClearEvent(&imageBitsEvid);
  if (imageXresEvid != NULL) viewImageCaClearEvent(&imageXresEvid);
  if (imageYresEvid != NULL) viewImageCaClearEvent(&imageYresEvid);
  if (imageSizeChid != NULL) viewImageCaClearChannel(&imageSizeChid);
  if (imageBitsChid != NULL) viewImageCaClearChannel(&imageBitsChid);
  if (imageXresChid != NULL) viewImageCaClearChannel(&imageXresChid);
  if (imageYresChid != NULL) viewImageCaClearChannel(&imageYresChid);
  viewImageCaPendIO(1.0);

  imageEvid=NULL;
  imageChid=NULL;
  imageSizeEvid=NULL;
  imageBitsEvid=NULL;
  imageXresEvid=NULL;
  imageYresEvid=NULL;
  imageSizeChid=NULL;
  imageBitsChid=NULL;
  imageXresChid=NULL;
  imageYresChid=NULL;
  imageUpdateChid=NULL;
  imageRoiXEvid=NULL;
  imageRoiYEvid=NULL;
  imageRoiWidthEvid=NULL;
  imageRoiHeightEvid=NULL;
  imageRoiXChid=NULL;
  imageRoiYChid=NULL;
  imageRoiWidthChid=NULL;
  imageRoiHeightChid=NULL;
  if (imageArrayChid) free(imageArrayChid);
  imageArrayChid = NULL;
  setRedLutWFChid=NULL;
  setBlueLutWFChid=NULL;
  setGreenLutWFChid=NULL;
  setRedLutWFEvid=NULL;
  setBlueLutWFEvid=NULL;
  setGreenLutWFEvid=NULL;
}

/******************************************************
  viewImageDataDisconnect
******************************************************/
void viewImageDataDisconnect()
{
  viewImageDataClearChannels();
  viewImageCaClose();
}

/******************************************************
  viewImageDataSizeValueEventCallback
******************************************************/
static void viewImageDataSizeValueEventCallback(struct event_handler_args args)
{    
  if ( args.status != ECA_NORMAL ) {
    printf ( "viewImageDataSizeValueEventCallback: failed because \"%s\"\n",
        ca_message ( args.status ) );
    return;
  }
  newImageArraySize = ((struct dbr_time_short *)args.dbr)->value;
  newImageArraySize++;
  imageModified=1;
}

/******************************************************
  viewImageDataBitsValueEventCallback
******************************************************/
static void viewImageDataBitsValueEventCallback(struct event_handler_args args)
{    
  if ( args.status != ECA_NORMAL ) {
    printf ( "viewImageDataBitsValueEventCallback: failed because \"%s\"\n",
        ca_message ( args.status ) );
    return;
  }
  newImageArrayBits = ((struct dbr_time_short *)args.dbr)->value;
  imageModified=1;
}

/******************************************************
  viewImageDataXresValueEventCallback
******************************************************/
static void viewImageDataXresValueEventCallback(struct event_handler_args args)
{    
  if ( args.status != ECA_NORMAL ) {
    printf ( "viewImageDataXresValueEventCallback: failed because \"%s\"\n",
        ca_message ( args.status ) );
    return;
  }
  newImageArrayWidth = ((struct dbr_time_short *)args.dbr)->value;
  imageModified=1;
}

/******************************************************
  viewImageDataYresValueEventCallback
******************************************************/
static void viewImageDataYresValueEventCallback(struct event_handler_args args)
{    
  if ( args.status != ECA_NORMAL ) {
    printf ( "viewImageDataYresValueEventCallback: failed because \"%s\"\n",
        ca_message ( args.status ) );
    return;
  }
  newImageArrayHeight = ((struct dbr_time_short *)args.dbr)->value;
  imageModified=1;
}

/******************************************************
  viewImageDataBitsConnectionEventCallback
******************************************************/
static void viewImageDataBitsConnectionEventCallback(struct connection_handler_args args)
{
  if (!viewImageCaIsConnected(imageBitsChid)) {
    errDialog("Not connected : %s%s\n",imageNamePrefix,imageBitsName);
  }else {
    if (!imageBitsEvid) {
      viewImageCaAddValueEvent(DBR_TIME_SHORT,imageBitsChid,&imageBitsEvid,viewImageDataBitsValueEventCallback);
      viewImageCaFlushIO();
    }
  }
}

/******************************************************
  viewImageDataSizeConnectionEventCallback
******************************************************/
static void viewImageDataSizeConnectionEventCallback(struct connection_handler_args args)
{
  if (!viewImageCaIsConnected(imageSizeChid)) {
    errDialog("Not connected : %s%s\n",imageNamePrefix,imageSizeName);
  }else {
    if (!imageSizeEvid) {
      viewImageCaAddValueEvent(DBR_TIME_SHORT,imageSizeChid,&imageSizeEvid,viewImageDataSizeValueEventCallback);
      viewImageCaFlushIO();
    }
  }
}

/******************************************************
  viewImageDataXresConnectionEventCallback
******************************************************/
static void viewImageDataXresConnectionEventCallback(struct connection_handler_args args)
{
  if (!viewImageCaIsConnected(imageXresChid)) {
    errDialog("Not connected : %s%s\n",imageNamePrefix,imageXresName);
  }else {
    if (!imageXresEvid) {
      viewImageCaAddValueEvent(DBR_TIME_SHORT,imageXresChid,&imageXresEvid,viewImageDataXresValueEventCallback);
      viewImageCaFlushIO();
    }
  }
}

/******************************************************
  viewImageDataYresConnectionEventCallback
******************************************************/
static void viewImageDataYresConnectionEventCallback(struct connection_handler_args args)
{
  if (!viewImageCaIsConnected(imageYresChid)) {
    errDialog("Not connected : %s%s\n",imageNamePrefix,imageYresName);
  }else {
    if (!imageYresEvid) {
      viewImageCaAddValueEvent(DBR_TIME_SHORT,imageYresChid,&imageYresEvid,viewImageDataYresValueEventCallback);
      viewImageCaFlushIO();
    }
  }
}

/******************************************************
  viewImageDataRoiXValueEventCallback
******************************************************/
static void viewImageDataRoiXValueEventCallback(struct event_handler_args args)
{    
  if ( args.status != ECA_NORMAL ) {
    printf ( "viewImageDataRoiXValueEventCallback: failed because \"%s\"\n",
        ca_message ( args.status ) );
    return;
  }
  imageRoiX = ((struct dbr_time_short *)args.dbr)->value;
  roiModified = 1;
}

/******************************************************
  viewImageDataRoiYValueEventCallback
******************************************************/
static void viewImageDataRoiYValueEventCallback(struct event_handler_args args)
{    
  if ( args.status != ECA_NORMAL ) {
    printf ( "viewImageDataRoiYValueEventCallback: failed because \"%s\"\n",
        ca_message ( args.status ) );
    return;
  }
  imageRoiY = ((struct dbr_time_short *)args.dbr)->value;
  roiModified = 1;
}

/******************************************************
  viewImageDataRoiWidthValueEventCallback
******************************************************/
static void viewImageDataRoiWidthValueEventCallback(struct event_handler_args args)
{    
  if ( args.status != ECA_NORMAL ) {
    printf ( "viewImageDataRoiWidthValueEventCallback: failed because \"%s\"\n",
        ca_message ( args.status ) );
    return;
  }
  imageRoiWidth = ((struct dbr_time_short *)args.dbr)->value;
  roiModified = 1;
}

/******************************************************
  viewImageDataRoiHeightValueEventCallback
******************************************************/
static void viewImageDataRoiHeightValueEventCallback(struct event_handler_args args)
{    
  if ( args.status != ECA_NORMAL ) {
    printf ( "viewImageDataRoiHeightValueEventCallback: failed because \"%s\"\n",
        ca_message ( args.status ) );
    return;
  }
  imageRoiHeight = ((struct dbr_time_short *)args.dbr)->value;
  roiModified = 1;
}

/******************************************************
  viewImageDataRoiXConnectionEventCallback
******************************************************/
static void viewImageDataRoiXConnectionEventCallback(struct connection_handler_args args)
{
  roiModified = 1;
  if (!viewImageCaIsConnected(imageRoiXChid)) {
    errDialog("Not connected : %s%s\n",imageNamePrefix,imageRoiXName);
  }else {
    if (!imageRoiXEvid) {
      viewImageCaAddValueEvent(DBR_TIME_SHORT,imageRoiXChid,&imageRoiXEvid,viewImageDataRoiXValueEventCallback);
      viewImageCaFlushIO();
    }
  }
}

/******************************************************
  viewImageDataRoiYConnectionEventCallback
******************************************************/
static void viewImageDataRoiYConnectionEventCallback(struct connection_handler_args args)
{
  roiModified = 1;
  if (!viewImageCaIsConnected(imageRoiYChid)) {
    errDialog("Not connected : %s%s\n",imageNamePrefix,imageRoiYName);
  }else {
    if (!imageRoiYEvid) {
      viewImageCaAddValueEvent(DBR_TIME_SHORT,imageRoiYChid,&imageRoiYEvid,viewImageDataRoiYValueEventCallback);
      viewImageCaFlushIO();
    }
  }
  return;
}

/******************************************************
  viewImageDataRoiWidthConnectionEventCallback
******************************************************/
static void viewImageDataRoiWidthConnectionEventCallback(struct connection_handler_args args)
{
  roiModified = 1;
  if (!viewImageCaIsConnected(imageRoiWidthChid)) {
    errDialog("Not connected : %s%s\n",imageNamePrefix,imageRoiWidthName);
  }else {
    if (!imageRoiWidthEvid) {
      viewImageCaAddValueEvent(DBR_TIME_SHORT,imageRoiWidthChid,&imageRoiWidthEvid,viewImageDataRoiWidthValueEventCallback);
      viewImageCaFlushIO();
    }
  }
}

/******************************************************
  viewImageDataRoiHeightConnectionEventCallback
******************************************************/
static void viewImageDataRoiHeightConnectionEventCallback(struct connection_handler_args args)
{
  roiModified = 1;
  if (!viewImageCaIsConnected(imageRoiHeightChid)) {
    errDialog("Not connected : %s%s\n",imageNamePrefix,imageRoiHeightName);
  }else {
    if (!imageRoiHeightEvid) {
      viewImageCaAddValueEvent(DBR_TIME_SHORT,imageRoiHeightChid,&imageRoiHeightEvid,viewImageDataRoiHeightValueEventCallback);
      viewImageCaFlushIO();
    }
  }
}

/******************************************************
  viewImageDataImageValueEventCallback
******************************************************/
static void viewImageDataImageValueEventCallback(struct event_handler_args args)
{    
  int stat, sevr;

  stat = ((struct dbr_sts_char *) args.dbr)->status;
  sevr = ((struct dbr_sts_char *) args.dbr)->severity;
  if (sevr>0 ){
    errDialog("WARNING: Alarm severity=%s: for %s\n",
    alarmSeverityString[sevr],ca_name(args.chid));
  }

  if (stat>0 ){
    errDialog("WARNING: Alarm status=%s: for %s\n",
    alarmStatusString[stat],ca_name(args.chid));
  }

  switch (args.status) {
  case ECA_NORMAL:
    imageModified=1;
#if 0
time_t now;
now=time(NULL);
printf ("ValueEventCallback:  %s",ctime(&now));
fflush(stdout);
#endif
    break;
  default:
    errDialog("value event failed: Return status: %s "
    " for %s\n",ca_message(args.status),ca_name(args.chid));
    break;
  }

}

/******************************************************
  viewImageDataImageConnectionEventCallback
******************************************************/
static void viewImageDataImageConnectionEventCallback(struct connection_handler_args args)
{
  if (viewImageCaIsConnected(imageChid)) {
    if (!imageEvid) {
      viewImageCaAddValueEvent(DBR_STS_CHAR,imageChid,&imageEvid,viewImageDataImageValueEventCallback);
      viewImageCaFlushIO();
    }
#if 0
  } else {
    errDialog("Not connected : %s%s\n",imageNamePrefix,imageName);
#endif
  }
}

/******************************************************
  viewImageDataSetRedLutWFValueEventCallback
******************************************************/
static void viewImageDataSetRedLutWFValueEventCallback(struct event_handler_args args)
{
  int arrayAllZeros=1;
  int i;
  int count;
  int type;
  struct dbr_sts_char *pvalue;
  unsigned char *pchar;

  if (args.status !=ECA_NORMAL) {
    errDialog("value event failed: Return status: %s "
      " for %s\n",ca_message(args.status),ca_name(args.chid));
    return;
  }

  count = args.count;
  type = args.type;
  pvalue = (void *)args.dbr;
  pchar = &pvalue->value;

  if (type!=DBR_STS_CHAR || count!=MAX_COLORS) {
    if (colorRedArray) free(colorRedArray);
    colorRedArray=0;
  } else {
    if (!colorRedArray) colorRedArray = (unsigned char *)calloc(MAX_COLORS,sizeof(unsigned char));
    /* dont use color array if all zeros */
    for (i = 0; i < MAX_COLORS; i++,pchar++) {
      colorRedArray[i]=*pchar;
      if (arrayAllZeros && (colorRedArray[i]!=0)) arrayAllZeros = 0;
    }
#if 0
printf ("\ncolorRedArray\n");
for (i = 0; i < MAX_COLORS; i++) {
  if(i>0 && i%22 == 0) printf("\n");
  printf ("%d ",(short)colorRedArray[i]);
}
printf ("\n\n");
#endif
    if (arrayAllZeros){
      free(colorRedArray);
      colorRedArray=0;
    }
  }
  colorTableModified=1;
}

/******************************************************
  viewImageDataSetBlueLutWFValueEventCallback
******************************************************/
static void viewImageDataSetBlueLutWFValueEventCallback(struct event_handler_args args)
{
  int arrayAllZeros=1;
  int i;
  int count;
  int type;
  struct dbr_sts_char *pvalue;
  unsigned char *pchar;

  count = args.count;
  type = args.type;
  pvalue = (void *)args.dbr;
  pchar = &pvalue->value;

  if (args.status !=ECA_NORMAL) {
    errDialog("value event failed: Return status: %s "
      " for %s\n",ca_message(args.status),ca_name(args.chid));
    return;
  }

  if (type!=DBR_STS_CHAR || count!=MAX_COLORS) {
    if (colorBlueArray) free(colorBlueArray);
    colorBlueArray=0;
  } else {
    if (!colorBlueArray) colorBlueArray = (unsigned char *)calloc(MAX_COLORS,sizeof(unsigned char));
    /* dont use color array if all zeros */
    for (i = 0; i < MAX_COLORS; i++,pchar++) {
      colorBlueArray[i]=*pchar;
      if (arrayAllZeros && (colorBlueArray[i]!=0)) arrayAllZeros = 0;
    }
#if 0
printf ("\ncolorBlueArray\n");
for (i = 0; i < MAX_COLORS; i++) {
  if(i>0 && i%22 == 0) printf("\n");
  printf ("%d ",(short)colorBlueArray[i]);
}
printf ("\n\n");
#endif
    if (arrayAllZeros){
      free(colorBlueArray);
      colorBlueArray=0;
    }
  }
  colorTableModified=1;
}

/******************************************************
  viewImageDataSetGreenLutWFValueEventCallback
******************************************************/
static void viewImageDataSetGreenLutWFValueEventCallback(struct event_handler_args args)
{
  int arrayAllZeros=1;
  int i;
  int count;
  int type;
  struct dbr_sts_char *pvalue;
  unsigned char *pchar;

  if (args.status !=ECA_NORMAL) {
    errDialog("value event failed: Return status: %s "
      " for %s\n",ca_message(args.status),ca_name(args.chid));
    return;
  }

  count = args.count;
  type = args.type;
  pvalue = (void *)args.dbr;
  pchar = &pvalue->value;

  if (type!=DBR_STS_CHAR || count!=MAX_COLORS) {
    if (colorGreenArray) free(colorGreenArray);
    colorGreenArray=0;
  } else {
    if (!colorGreenArray) colorGreenArray = (unsigned char *)calloc(MAX_COLORS,sizeof(unsigned char));
    /* dont use color array if all zeros */
    for (i = 0; i < MAX_COLORS; i++,pchar++) {
      colorGreenArray[i]=*pchar;
      if (arrayAllZeros && (colorGreenArray[i]!=0)) arrayAllZeros = 0;
    }
#if 0
printf ("\ncolorGreenArray\n");
for (i = 0; i < MAX_COLORS; i++) {
  if(i>0 && i%22 == 0) printf("\n");
  printf ("%d ",(short)colorGreenArray[i]);
}
printf ("\n\n");
#endif
    if (arrayAllZeros){
      free(colorGreenArray);
      colorGreenArray=0;
    }
  }
  colorTableModified=1;
}

/******************************************************
  viewImageDataSetRedLutWFConnectionEventCallback
******************************************************/
static void viewImageDataSetRedLutWFConnectionEventCallback(void *unused)
{
  if (!viewImageCaIsConnected(setRedLutWFChid)) {
    errDialog("Not connected : %s%s\n",imageNamePrefix,setRedLutWFName);
    if (colorRedArray) free(colorRedArray);
    colorRedArray=0;
    colorTableModified=1;
  }else {
    if (!setRedLutWFEvid) {
      viewImageCaAddValueEvent(DBR_STS_CHAR,setRedLutWFChid,&setRedLutWFEvid,
        viewImageDataSetRedLutWFValueEventCallback);
      viewImageCaFlushIO();
    }
  }
  return;
}

/******************************************************
  viewImageDataSetBlueLutWFConnectionEventCallback
******************************************************/
static void viewImageDataSetBlueLutWFConnectionEventCallback(void *unused)
{
  if (!viewImageCaIsConnected(setBlueLutWFChid)) {
    errDialog("Not connected : %s%s\n",imageNamePrefix,setBlueLutWFName);
    if (colorBlueArray) free(colorBlueArray);
    colorBlueArray=0;
    colorTableModified=1;
  }else {
    if (!setBlueLutWFEvid) {
      viewImageCaAddValueEvent(DBR_STS_CHAR,setBlueLutWFChid,&setBlueLutWFEvid,
        viewImageDataSetBlueLutWFValueEventCallback);
      viewImageCaFlushIO();
    }
  }
  return;
}

/******************************************************
  viewImageDataSetGreenLutWFConnectionEventCallback
******************************************************/
static void viewImageDataSetGreenLutWFConnectionEventCallback(void *unused)
{
  if (!viewImageCaIsConnected(setGreenLutWFChid)) {
    errDialog("Not connected : %s%s\n",imageNamePrefix,setGreenLutWFName);
    if (colorGreenArray) free(colorGreenArray);
    colorGreenArray=0;
    colorTableModified=1;
  }else {
    if (!setGreenLutWFEvid) {
      viewImageCaAddValueEvent(DBR_STS_CHAR,setGreenLutWFChid,&setGreenLutWFEvid,
        viewImageDataSetGreenLutWFValueEventCallback);
      viewImageCaFlushIO();
    }
  }
  return;
}

/******************************************************
  viewImageDataConnect
******************************************************/
void viewImageDataConnect(char *imageNamePrefixIn)
{
  imageNamePrefix=imageNamePrefixIn;
  if (neverConnected) viewImageCaInit();

  sprintf(name,"%s%s",imageNamePrefix,imageName);
  viewImageCaSearchAndConnectChannel(name,&imageChid,viewImageDataImageConnectionEventCallback);
  sprintf(name,"%s%s",imageNamePrefix,imageSizeName);
  viewImageCaSearchAndConnectChannel(name,&imageSizeChid,viewImageDataSizeConnectionEventCallback);
  sprintf(name,"%s%s",imageNamePrefix,imageBitsName);
  viewImageCaSearchAndConnectChannel(name,&imageBitsChid,viewImageDataBitsConnectionEventCallback);
  sprintf(name,"%s%s",imageNamePrefix,imageXresName);
  viewImageCaSearchAndConnectChannel(name,&imageXresChid,viewImageDataXresConnectionEventCallback);
  sprintf(name,"%s%s",imageNamePrefix,imageYresName);
  viewImageCaSearchAndConnectChannel(name,&imageYresChid,viewImageDataYresConnectionEventCallback);
  sprintf(name,"%s%s",imageNamePrefix,imageUpdateName);
  viewImageCaSearchChannel(name,&imageUpdateChid);
  sprintf(name,"%s%s",imageNamePrefix,imageRoiXName);
  viewImageCaSearchAndConnectChannel(name,&imageRoiXChid,viewImageDataRoiXConnectionEventCallback);
  sprintf(name,"%s%s",imageNamePrefix,imageRoiYName);
  viewImageCaSearchAndConnectChannel(name,&imageRoiYChid,viewImageDataRoiYConnectionEventCallback);
  sprintf(name,"%s%s",imageNamePrefix,imageRoiWidthName);
  viewImageCaSearchAndConnectChannel(name,&imageRoiWidthChid,viewImageDataRoiWidthConnectionEventCallback);
  sprintf(name,"%s%s",imageNamePrefix,imageRoiHeightName);
  viewImageCaSearchAndConnectChannel(name,&imageRoiHeightChid,viewImageDataRoiHeightConnectionEventCallback);
  sprintf(name,"%s%s",imageNamePrefix,setRedLutWFName);
  viewImageCaSearchAndConnectChannel(name,&setRedLutWFChid,viewImageDataSetRedLutWFConnectionEventCallback);
  sprintf(name,"%s%s",imageNamePrefix,setBlueLutWFName);
  viewImageCaSearchAndConnectChannel(name,&setBlueLutWFChid,viewImageDataSetBlueLutWFConnectionEventCallback);
  sprintf(name,"%s%s",imageNamePrefix,setGreenLutWFName);
  viewImageCaSearchAndConnectChannel(name,&setGreenLutWFChid,viewImageDataSetGreenLutWFConnectionEventCallback);

  viewImageCaPendIO(5.0);

  imageArrayBits=0;
  viewImageCaArrayGet(DBR_LONG,1,imageXresChid,&imageArrayWidth);
  viewImageCaArrayGet(DBR_LONG,1,imageYresChid,&imageArrayHeight);
  viewImageCaArrayGet(DBR_LONG,1,imageSizeChid,&imageArraySize);
  viewImageCaArrayGet(DBR_LONG,1,imageBitsChid,&imageArrayBits);
  viewImageCaArrayGet(DBR_LONG,1,imageRoiXChid,&imageRoiX);
  viewImageCaArrayGet(DBR_LONG,1,imageRoiYChid,&imageRoiY);
  viewImageCaArrayGet(DBR_LONG,1,imageRoiWidthChid,&imageRoiWidth);
  viewImageCaArrayGet(DBR_LONG,1,imageRoiHeightChid,&imageRoiHeight);

  viewImageCaPendIO(5.0);

  if (!viewImageCaIsConnected(imageChid))
        errDialog("Not connected : %s\n",imageNamePrefix,imageName);
  if (!viewImageCaIsConnected(imageSizeChid))
        errDialog("Not connected : %s%s\n",imageNamePrefix,imageSizeName);
  /* It is OK if *image.BITS pv does not exist! */
  /*if (!viewImageCaIsConnected(imageBitsChid)) */
        /*errDialog("Not connected : %s%s\n",imageNamePrefix,imageBitsName); */
  if (!viewImageCaIsConnected(imageXresChid))
        errDialog("Not connected : %s%s\n",imageNamePrefix,imageXresName);
  if (!viewImageCaIsConnected(imageYresChid))
        errDialog("Not connected : %s%s\n",imageNamePrefix,imageYresName);
  if (!viewImageCaIsConnected(imageUpdateChid))
        errDialog("Not connected : %s%s\n",imageNamePrefix,imageUpdateName);
  if (!viewImageCaIsConnected(imageRoiXChid))
        errDialog("Not connected : %s%s\n",imageNamePrefix,imageRoiXName);
  if (!viewImageCaIsConnected(imageRoiYChid))
        errDialog("Not connected : %s%s\n",imageNamePrefix,imageRoiYName);
  if (!viewImageCaIsConnected(imageRoiWidthChid))
        errDialog("Not connected : %s%s\n",imageNamePrefix,imageRoiWidthName);
  if (!viewImageCaIsConnected(imageRoiHeightChid))
        errDialog("Not connected : %s%s\n",imageNamePrefix,imageRoiHeightName);
  if (!viewImageCaIsConnected(setRedLutWFChid))
        errDialog("Not connected : %s%s\n",imageNamePrefix,setRedLutWFName);
  if (!viewImageCaIsConnected(setBlueLutWFChid))
        errDialog("Not connected : %s%s\n",imageNamePrefix,setBlueLutWFName);
  if (!viewImageCaIsConnected(setGreenLutWFChid))
        errDialog("Not connected : %s%s\n",imageNamePrefix,setGreenLutWFName);

  imageArraySize++;

  neverConnected=0;
}

/******************************************************
  viewImageDataCreateData
******************************************************/
void viewImageDataCreateData(char *imageNamePrefix)
{
    viewImageDataClearChannels();
    viewImageDataSetRoi(0,0,0,0);
    if (imageNamePrefix) viewImageDataConnect(imageNamePrefix);

    imageModified=1;
}

/******************************************************
  viewImageDataReadImage
******************************************************/
int viewImageDataReadImage()
{
  int i,imageMod=0;
  int arrayDBR,arrayWidth;

#if 0
readImageCount++;
#endif

#if 0
clock_t start=0,finish;
double duration;
start=clock();
#endif

  if (imageModified) imageMod=1;

  if (viewImageCaIsConnected(imageChid)) {

    if ( newImageArraySize!=imageArraySize ||
         newImageArrayWidth!=imageArrayWidth ||
         newImageArrayHeight!=imageArrayHeight ) {
      if ( newImageArrayWidth*newImageArrayHeight*newImageArraySize > 
           imageArrayHeight*imageArrayWidth*imageArraySize ) {
        if (imageArray ) free(imageArray);
        imageArray = (unsigned char *)calloc(newImageArrayWidth*newImageArrayHeight*newImageArraySize,sizeof(unsigned char));
      }
      if ( newImageArrayHeight != imageArrayHeight || imageArrayChid==NULL) {
        if (imageArrayHeight>0 && imageArrayChid) {
          for (i=0;i<imageArrayHeight;i++) {
            if (imageArrayChid[i]!=0) viewImageCaClearChannel(&imageArrayChid[i]);
          }
          viewImageCaPendIO(5.0);
        }
        if (imageArrayChid) free(imageArrayChid);
        imageArrayChid = NULL;
        if (newImageArrayHeight>0) {
          imageArrayChid = (chid *)calloc(newImageArrayHeight,sizeof(chid));
          for (i=0;i<newImageArrayHeight;i++) {
            sprintf(name,"%s%s%d",imageNamePrefix,imageArrayName,i);
            viewImageCaSearchChannel(name,&imageArrayChid[i]);
          }
          viewImageCaPendIO(5.0);
        }
      }
      imageArrayWidth=newImageArrayWidth;
      imageArrayHeight=newImageArrayHeight;
      imageArraySize=newImageArraySize;
      imageMod=1;
    }
    if (!imageArray) {
      if (imageArray ) free(imageArray);
      imageArray = (unsigned char *)calloc(newImageArrayWidth*newImageArrayHeight*newImageArraySize,sizeof(unsigned char));
      if (imageArrayChid) free(imageArrayChid);
      imageArrayChid = NULL;
      if (newImageArrayHeight>0) {
        imageArrayChid = (chid *)calloc(newImageArrayHeight,sizeof(chid));
        for (i=0;i<newImageArrayHeight;i++) {
          sprintf(name,"%s%s%d",imageNamePrefix,imageArrayName,i);
          viewImageCaSearchChannel(name,&imageArrayChid[i]);
        }
        viewImageCaPendIO(5.0);
      }
      imageArrayWidth=newImageArrayWidth;
      imageArrayHeight=newImageArrayHeight;
      imageArraySize=newImageArraySize;
      imageMod=1;
    }
  } else {
    if (imageArray ) free(imageArray);
    imageArray=0;
    imageMod = 1;
  }

  if (!viewImageDataRoiIsConnected()){
    imageRoiX=0;
    imageRoiY=0;
    imageRoiWidth=0;
    imageRoiHeight=0;
  }

  if (imageMod || subRegionOnlyModified || (roiModified && subRegionOnly) ) {
    imageX=0;
    imageY=0;
    imageWidth=imageArrayWidth;
    imageHeight=imageArrayHeight;
    if (subRegionOnly) {
      if (imageSubX >= 0) imageX=imageSubX;
      if (imageSubY >= 0) imageY=imageSubY;
      if (imageSubWidth > 0) imageWidth=imageSubWidth;
      if (imageSubHeight > 0) imageHeight=imageSubHeight;
    }
  }

  if (imageMod && viewImageCaIsConnected(imageChid)) {
    if ( imageHeight>0 && imageArrayWidth>0 ) {

      arrayWidth=imageArrayWidth*imageArraySize;

      /* CA gets for displayed rows */
      arrayDBR=imageArrayDBR[imageArraySize-1];
      for (i=imageY; i < imageHeight+imageY; i++) {
        viewImageCaArrayGet(arrayDBR,0,imageArrayChid[i],&imageArray[arrayWidth*i]);
      }
      viewImageCaPendIO(5.0);

    }
  }
  imageModified=0;
  roiModified = 0;
  subRegionOnlyModified = 0;

#if 0
finish = clock();
duration = finish-start;
duration = duration/CLOCKS_PER_SEC;
now=time(NULL);
printf("viewImageDataReadImage took                %15.10f total sec  %s", duration,ctime(&now));
fflush(stdout);
#endif

  return 0;
}

/******************************************************
  viewImageDataUpdateArrayMinMax()
******************************************************/
int viewImageDataUpdateArrayMinMax()
{
  int i,j,k,value;
  int arrayWidth;
  
      imageMinValue=0;
      imageMaxValue=0;
      if (!imageArray ) return -1;

      /* set initial min and max values */
      arrayWidth = imageArrayWidth*imageArraySize;
      i=arrayWidth*imageY + imageArraySize*imageX;
      if (imageArraySize==2) imageMinValue = *((unsigned short *)(&imageArray[i]));
      else value = imageMinValue = imageArray[i];
      imageMaxValue=0;

      /* traverse displayed imageArray to determine min and max */
      for (i=imageY; i < imageHeight; i++) {
        for (j=imageX; j < imageWidth; j++) {
          k=arrayWidth*i + imageArraySize*j; 
          if (imageArraySize==2) value = *((unsigned short *)(&imageArray[k]));
          else value = imageArray[k];
          if (value<imageMinValue) imageMinValue=value;
          if (value>imageMaxValue) imageMaxValue=value;
#if 0
if (k < 5000) printf("min=%d max=%d value=%ld   imageArray[%d]=%ld %ld \n",imageMinValue,imageMaxValue,value,k,imageArray[k],imageArray[k+1]);
printf ("imageArraySize=%d i=%d arrayWidth=%d imageY=%d  imageX=%d\n",imageArraySize,i,arrayWidth,imageY,imageX);
#endif
        }
      }
/*
printf (" imageMinValue=%d  imageMaxValue=%d \n",imageMinValue,imageMaxValue );
*/
  return 0;
}

