/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/* viewImageErrDialog.c */

/* viewImageErrDialog.c,v 1.4 2002/07/23 19:18:24 jba Exp */
/************************DESCRIPTION***********************************
  Creates  error message popup dialog
**********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include <Xm/Xm.h>
#include <Xm/MessageB.h>
#include <Xm/FileSB.h>

#include "viewImageErrDialog.h"

/* globals */
extern Display *display;
extern Widget topLevelShell;

/* locals */
static Widget warningbox = NULL;

/* forward definitions */
static void okCallback(Widget w, XtPointer clientdata, XtPointer calldata);

/******************************************************
  Error message popup dialog
******************************************************/
void errDialog(const char *fmt, ...)
{
	XmString cstring,cstringOld,cstringNew;
	va_list vargs;
	static char lstring[2000];  /* DANGER: Fixed buffer size */
	static int warningboxMessages = 0;
	int nargs=10;
	Arg args[10];
    Dimension shellHeight, shellWidth;
    Position newY, newX;
	char warningboxName[]="warningMessage";
	char warningboxTitle[]="viewImage: Warning Messages"; 

	va_start(vargs,fmt);
	vsprintf(lstring,fmt,vargs);
	va_end(vargs);

	if(lstring[0] == '\0') return;

	if (warningbox) {

      if (XtIsManaged(warningbox)){

		if (warningboxMessages > 30) return;

		cstring=XmStringCreateLtoR(lstring,XmSTRING_DEFAULT_CHARSET);
		XtVaGetValues(warningbox, XmNmessageString, &cstringOld, NULL);
		cstringNew = XmStringConcat(cstringOld,cstring);
		XmStringFree(cstring);
		XmStringFree(cstringOld);
		if (warningboxMessages == 30){
			cstring=XmStringCreateLtoR(
				"\nOnly first 30 messages are displayed\n",
				XmSTRING_DEFAULT_CHARSET);
			cstringOld = cstringNew;
			cstringNew = XmStringConcat(cstringOld,cstring);
			XmStringFree(cstring);
			XmStringFree(cstringOld);
		}
		XtVaSetValues(warningbox, XmNmessageString, cstringNew, NULL);
		XmStringFree(cstringNew);
		warningboxMessages += 1;
		XtManageChild(warningbox);
	  } else {
		cstringNew=XmStringCreateLtoR(lstring,XmSTRING_DEFAULT_CHARSET);
		XtVaSetValues(warningbox, XmNmessageString, cstringNew, NULL);
		XmStringFree(cstringNew);
		warningboxMessages = 1;
		XtManageChild(warningbox);
      }
	} else {
		XBell(display,50); 
		XBell(display,50); 
		XBell(display,50);
		cstring=XmStringCreateLtoR(lstring,XmSTRING_DEFAULT_CHARSET);
		nargs=0;
		XtSetArg(args[nargs],XmNtitle,warningboxTitle); 
		nargs++;
		XtSetArg(args[nargs],XmNmessageString,cstring); 
		nargs++;

		warningbox=XmCreateWarningDialog(topLevelShell,warningboxName,
		    args,nargs);

        XtVaGetValues(topLevelShell,
          XmNheight,&shellHeight,
          XmNwidth, &shellWidth,
          NULL);
        newX=shellWidth/2;
        newY=shellHeight/2;
    
        XtVaSetValues(warningbox,
          XmNdefaultPosition, False,
          XmNx,               newX,
          XmNy,               newY,
          NULL);

		XmStringFree(cstring);
		XtDestroyWidget(XmMessageBoxGetChild(warningbox,XmDIALOG_CANCEL_BUTTON));
		XtDestroyWidget(XmMessageBoxGetChild(warningbox,XmDIALOG_HELP_BUTTON));
		XtAddCallback(warningbox,XmNokCallback,okCallback,NULL);
		warningboxMessages = 1;
		XtManageChild(warningbox);
	}
}

/******************************************************
  Static callback routine for exiting application
******************************************************/
static void exitQuit(Widget w, XtPointer clientdata, XtPointer calldata)
{
	exit(1);
}

/******************************************************
  Fatal error message popup dialog
******************************************************/
void fatalErrDialog(const char *fmt, ...)
{
	va_list vargs;
	static char lstring[1024];  /* DANGER: Fixed buffer size */

	va_start(vargs,fmt);
	vsprintf(lstring,fmt,vargs);
	va_end(vargs);

	if(lstring[0] == '\0') return;

	errDialog(lstring);
	if (warningbox) XtAddCallback(warningbox,XmNokCallback,exitQuit,NULL);
}

/******************************************************
  Static callback routine for destroying a widget 
******************************************************/
static void okCallback(Widget w, XtPointer clientdata, XtPointer calldata)
{
    XtUnmanageChild(w);
}
/******************************************************
  Static callback routine for destroying a widget 
******************************************************/
void errDialogRaiseWindow()
{
	if (warningbox != 0) {
        XRaiseWindow(display, XtWindow(XtParent(warningbox)));
    }
}
